// AIvan.c 
// the AI of Ivan my compainion robot 
// top level c program file
// 
// by Les Hall
// started Thu Aug 18, 2016, 3:24 PM
// 



// include header files
#include <stdlib.h>  // utility functions
#include <stdio.h>  // the standard input / output header
#include <math.h>  // math functions
#include <time.h>  // time functions
#include "AIvan.h"  // Ivan's header file



// define structures
// 
// define the meta data structure
struct metaData {
  int days, awake, asleep;
};
// 
// "Sometimes a cigar is just a cigar." - Freud
struct sigmoid {
  float x;  // the input value
  float tau;  // scales x
  float shape;  // 1.0 for flat, 0.1 fo sharp, 0.01 for squareish
  float y;  // the output value
};
// 
// define the input/output audio and video stream buffers
struct input_output {
  int video[SIZE_X][SIZE_Y];
  float audio[AUDIO_BUFFER_SIZE];
};
// 
// define the cell structure
struct cell {
  float weights[SIZE_W];  // work like neural net weights
  float deltas[SIZE_W];  // most recent change in weights
  float sum;  // sum of weighted inputs and previous values
  float threshold;  // when sum exceeds this value, cell triggers
  float storage;  // stores output to act as latch (prevents ***dataCreep***)
  float output;  // stores the triggered output value for one cycle
};
// 
// define the block structure
struct shape {
  struct cell cells[SIZE_C];  // declare the cells in a block
};
// 
// define the array structure
struct cognitiveArray {
  struct shape shapes[SIZE_X][SIZE_Y][SIZE_Z];  // the array of cells
  double activity;
  double tau;
  float temperature;
  float fom;
  float command;
  struct sigmoid sig;
};




// declare functions
void switch_on(FILE *fp, struct metaData *md, struct cognitiveArray *ca);  // ON
void wake_up();  // wake Ivan up
void think(FILE *fp);  // Ivan thinks like Rodin's "Thinking Man"
int imagine();  // Ivan is daydreaming (dreaming while awake)
void go_to_sleep();  // tell Ivan to go to sleep
void dream();  // Ivan is dreaming
void switch_off(FILE *fp);  // switch Ivan off
void print_mode();  // print Ivan's mode to stdout
void get_input(struct input_output *is, struct cognitiveArray *ca);
void iterate(struct cognitiveArray *ca, 
  struct input_output *input, struct input_output *output, 
  int mode); // compute one iteration of calculations
void get_output(struct input_output *os, struct cognitiveArray *ca);
void diaplay(struct input_output *s);
void display_slice(int Z, struct cognitiveArray *ca);
void display_array(struct cognitiveArray *ca);
void init(struct cognitiveArray *ca);
int min(int a, int b);
int max(int a, int b);
void deltas(struct cognitiveArray *ca, int mode);
void fom(struct cognitiveArray *ca, int mode);
void sigmoid(struct sigmoid *s);



// the top level function
int main(int argc, char *argv[]) { 

  // declare variables
  struct metaData meta;  // declare the meta data structure
  struct cognitiveArray ca;  // declare the array of block
  struct input_output input;  // holds input video and audio
  struct input_output output;  // holds output video and audio
  FILE *fp;  // file pointer to Ivan's long term memory
  int mode = off; // Ivan's mode of operation, defined in AIvan.h


  // seed the random number generator with time()
  srand(time(0));

  // switch Ivan on - read in long term memory from disk file
  // and launch all of Ivan's cellular processes
  switch_on(fp, &meta, &ca);
  int dayCounter = meta.days;
  init(&ca);  // initialize variables in the cognitive array


  // keep running loop as long as Ivan's on cycle counter is positive
  while (dayCounter > 0) {

    // reset counters
    int wakingCounter = meta.awake;
    int sleepingCounter = meta.asleep;

    // Ivan wakes up for another day
    wake_up();  // send a wake-up signal to all of Ivan's cellular processes
    
    
    // keep runing loop as long as Ivan's day cycle counter is positive
    while(wakingCounter  > 0) {

      // print waking counter
      printf("wake Cycle:  %d", wakingCounter);

      mode = awake;  // set Ivan's mode to awake

      // Ivan's daytime mode
      if (imagine() ) {  // if Ivan is using his imagination

        mode |= dreaming;  // set the mode accordingy

      }

      // generate weight deltas for training purposes
      // and add them to the weights with mutiplied values according to the fom() function
      deltas(&ca, mode);
      
      // compute one iteration of AIvan's cognitive array
      iterate(&ca, &input, &output, mode); 

      // evaluate figure of merit (greater values are better)
      fom(&ca, mode);

      // remove deltas if poor figure of merit encountered
      // note:  get temperature and rand() involved in a somewhat
      //        analagous way to simulated annealing (SA)
      //        (but not theoretically exactly SA)

      // display the cognitive array outputs
      printf("\tactivity level:  %e", ca.activity);
      printf("\ttemperature:  %e", ca.temperature);
      display_array(&ca);

      // tidy up chores
      --wakingCounter; // decrement the day cycle counter

    }

    
    // go to sleep Ivan
    go_to_sleep();  // any chores in the process of going to sleep go here

    // sleep as long as Ivan's night cycle counter is positive
    while(sleepingCounter > 0) {
  
      // determine Ivan's mode for this cycle of the night
      mode = asleep;  // set Ivan's mode to asleep
      mode |= dreaming; // dream big Ivan, dream big

      // only dream if the dream flag is on in Ivan's mode
      if (mode & dreaming) {

        dream();  // tell Ivan's cellular processses to dream

      }

      // decrement the on cycle counter
      --sleepingCounter;
    
    }
  
    // decrement the on cycle counter
    --dayCounter;
  }
      

  // switch Ivan off
  switch_off(fp);  

  // end of program
  exit(0);

}



// switch Ivan on
void switch_on(FILE *fp, struct metaData *md, struct cognitiveArray *ca) {

  // define local variables
  int days, awake, asleep;
  float threshold;

  // read in Ivan's disk file (his long term memory)
  if ((fp = fopen(fileName, fileMode)) == NULL) {  // open file and check for error`

    printf("error opening file");

  } else {

    // scan in the meta data
    fscanf(fp, "%d %d %d\n", &days, &awake, &asleep);  
    fscanf(fp, "threshold %f\n", &threshold);  
    
    // assign the scanned data
    md->days = days;
    md->awake = awake;
    md->asleep = asleep;

    for(int z = 0; z < SIZE_Z; ++z) {  // loop thru z
  
      for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y

        for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

          for(int c = 0; c < SIZE_C; ++c) {  // loop thru c

            // copy the threshold value into the cognitive array
            ca->shapes[x][y][z].cells[c].threshold = threshold;

          }

        }

      }

    }

  }

  // print switch on message
  printf("\n\n\n\nIvan switched on!  \n");

}



// wake Ivan up
void wake_up() {

  // print welcome message
  printf("\tgood morning!  \n");

}



// do things while awake
void think() {
  
  // write out Ivan's state of mind
  printf("\t\tIvan is thinking\n");

}



// determine if Ivan is exercising his imagination
int imagine() {

  return ( rand() < (RAND_MAX/4) ) ? 1 : 0;

}




// tell Ivan to go to sleep
void go_to_sleep() {

  // print going to sleep message
  printf("\tgood night!\n");

}



// Ivan dreams at night
void dream() {

  // print going to sleep message
  printf("\t\tIvan is dreaming\n");

}



// turn off Ivan
void switch_off(FILE *fp) {

  // final print message
  printf("Ivan is switching off\n");

  // close Ivan's long term memory
  if ( fclose(fp) == EOF ) {


  } else {

    printf("error:  error closing file\n");

  }

  // final print message
  printf("Ivan is switched off\n\n\n\n");

}



// print Ivan's mode
void print_mode(int mode) {

  // begin with intro text
  printf("Ivan's mode is");

  // select modes with separate conditionals
  if (mode < 0)
    printf(" error");
  if (mode & off)
    printf(" off");
  if (mode & awake)
    printf(" awake");
  if (mode & asleep)
    printf(" asleep");
  if (mode & dreaming)
    printf(" dreaming");
  if (mode & imagining)
    printf(" imagining");

  // end with a newline
  printf("\n");

}



// get the input data stream
void get_input(struct input_output *is, struct cognitiveArray *ca) {

  // loop thru the array and fill in some values
  for (int y = 0; y < SIZE_Y; ++y) {  // loop thru y

    // calculate the Y dimension sine value
    float ySin = (float)sin(2.0*3.14159 * (double)y / (double)SIZE_Y);

    for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

      // calculate the X dimension sine value
      float xSin = (float)sin(2.0*3.14159 * (double)x / (double)SIZE_X);

      // add xVal times yVal to make a wavy display
      float val = 5.0 * (xSin + ySin);

      for(int c = 0; c < SIZE_C; ++c) {  // loop thru c
  
        // put the display value into both 
        // the output and the storage locations
        // of the cognitive array data structure
        ca->shapes[x][y][0].cells[c].output = val;  // output value
        ca->shapes[x][y][0].cells[c].storage = val;  // storage value

      }
  
    }

  }
  
  // create an audio stream
  for(int i = 0; i < AUDIO_BUFFER_SIZE; ++i) {

    is->audio[i] = (float)sin(2.0*3.14159 * (double)i / (double)AUDIO_BUFFER_SIZE);

  }

}



// compute one iteration of the array!
void iterate(struct cognitiveArray *ca, 
  struct input_output *input, struct input_output *output, 
  int mode) {

  // print the message that Ivan is thinking
  printf("\tIvan is thinking");
  
  // run the calculations for one iteration
  // and do not shift the storage value to the output just yet
  // loop thru the array
  //printf("running one iteration of calculations (sum of products)\n");
  for(int z = 0; z < SIZE_Z; ++z) {  // loop thru z

    for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y

      for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

        for(int c = 0; c < SIZE_C; ++c) {  // loop thru c

         for(int w = 0; w < SIZE_W; ++w) {  // loop thru w

           // use a case statement to convert the w value to a neighboring direction
           int X = x;
           int Y = y;
           int Z = z;
           switch (w) {
             case 1:  X = x-1;	break;
             case 2:  X = x+1;	break;
             case 3:  Y = y-1;  break;
             case 4:  Y = y+1;	break;
             case 5:  Z = z-1;	break;
             case 6:  Z = z+1;	break;
             default:  X = x, Y = y, Z = z;
           }

            //wrap excess values around with modulus operators
            if (X<0) X += SIZE_X;
            if (Y<0) Y += SIZE_Y;
            if (Z<0) Z += SIZE_Z;
            X %= SIZE_X;
            Y %= SIZE_Y;
            Z %= SIZE_Z;

            // multiply the neighbor index provided by the case statement 
            // by the corresponding weight index and add it to this cell's sum
            // (including possible small sums from previous iterations)
            ca->shapes[x][y][z].cells[c].sum += 
              ca->shapes[x][y][z].cells[c].weights[w] * 
              ca->shapes[X][Y][Z].cells[c].output;
 
          }

          // so now see if the threshold has been reached and 
          // fire the neuron if the threshold has been reached or exceeded
          if ( (ca->shapes[x][y][z].cells[c].sum) >= (ca->shapes[x][y][z].cells[c].threshold) ) {

            // copy the sum to the storage value
            ca->shapes[x][y][z].cells[c].storage -= ca->shapes[x][y][z].cells[c].sum;


            // erase the sum value because it has transferred to the storage value
            ca->shapes[x][y][z].cells[c].sum = 0.0;


            // for now just let the fired value sit in storage.  
            // it will get transferred to the output in the next loop set
            // this way we don't have data creeping in every direction

          }

        }

      }

    }

  }


  // apply the input data steam to Ivan's cognitive array
  get_input(input, ca);

  double total = 0.0;  // create an activity total

  // now shift all of the storage values to the output values // loop thru the array
  for(int z = 0; z < SIZE_Z; ++z) {  // loop thru z

    for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y

      for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

        for(int c = 0; c < SIZE_C; ++c) {  // loop thru c

          // calculate the sigmoid of the stored value
          ca->sig.x = ca->shapes[x][y][z].cells[c].storage;
          sigmoid(&(ca->sig));

          // transfer the storage value to the output
          ca->shapes[x][y][z].cells[c].output = ca->sig.y;

          // add the output value to the activity total
          total += fabs( ((double)ca->shapes[x][y][z].cells[c].output) );

          // erase the storage value
          ca->shapes[x][y][z].cells[c].storage = 0.0;

        }

      }

    }

  }

  // calculate size of array
  double arraySize = SIZE_X * SIZE_Y * SIZE_Z * SIZE_C;

  // low pass filter the running average of the activity total
  ca->activity = (ca->tau) * (ca->activity)  +  (1.0 - (ca->tau) ) * total / (arraySize);

}



// get the output data stream
void get_output(struct input_output *os, struct cognitiveArray *ca) {

  // and finally update the output values
  // deliver video output patterns to output structure
  // loop thru the array
  for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y
  
    for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

      os->video[x][y] = ca->shapes[x][y][SIZE_Z-1].cells[6].output;  

    }

  }
  // 
  // apply audio output buffer
  for(int i = 0; i < AUDIO_BUFFER_SIZE; ++i) {

    // just a null output for now until I figure out the mapping
    // between the cognitive array and the audio buffers 
    os->audio[i] = 0; 

  }
  
}



// display a video stream
void diaplay(struct input_output *s) {

  // determine the maximum brightness for scaling
  float max = 0;

  // loop thru the array
  for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y
  
    for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

      float val = s->video[x][y];
      if (val > max)
        max = val;

    }

  }

  printf("\n\n");

  // loop thru the array
  for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y
  
    for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

      int d = (int) ( 8 * s->video[x][y] / max );
      printf("%c", '1' + d);

    }
    
    printf("\n");

  }

  printf("\n\n");

}



void display_slice(int Z, struct cognitiveArray *ca) {

  // determine the maximum brightness for scaling
  float max = 0;

  // loop thru the array
  for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y
  
    for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

      float val = ca->shapes[x][y][Z].cells[6].output;
      if (val > max)
        max = val;

    }

  }

  printf("\n\n");

  // loop thru the array
  for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y
  
    for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

      int d = (int) ( 8 * ca->shapes[x][y][Z].cells[6].output / max );
      printf("%c", '1' + d);

    }
    
    printf("\n");

  }

  printf("\n\n");

}



void display_array(struct cognitiveArray *ca) {

  // determine the maximum brightness for scaling
  float min = 0;
  float max = 0;
  const char *ASCII = " .:-=+*#%@";

  // loop thru the array
  for(int z = 0; z < SIZE_Z; ++z) {  // loop thru z

    for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y
  
      for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

        float val = ca->shapes[x][y][z].cells[6].output;
        if (val < min) min = val;
        if (val > max) max = val;

      }

    }

  }

  float range = max - min;

  printf("\n\n");

  // make sure no div by 0 causes a segmentation fault:  11
  if (range > 0) {

    // loop thru the array
    for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y

      for(int z = 0; z < SIZE_Z; ++z) {  // loop thru z
  
        for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

          int d = (int) ( 4 * fabsf(ca->shapes[x][y][z].cells[6].output) / range );

          if ( (d <= 9) && (d >= 0) ) {

            printf("%c", ASCII[d]);

          } else {

            printf("!");

          }

        }

        printf(" ");

      }
    
      printf("\n");

    }

    printf("\n\n");

  }

}



// initialize variables
void init(struct cognitiveArray *ca) {

  // sigmoid values inside of cognitive array
  ca->sig.tau = 1.0;
  ca->sig.shape = 1.0;

  // cognitive array
  ca->tau = 0.95; // initialize time constant
  ca->activity = 0.1;  // initialize the activity level
  ca->temperature = 0.1; // initialize the temperature
  ca->command = 1.0;  // deltas command multiplier
  ca->fom = 1.0;  // figure of merit initial value
  
  // initialize weights
  // loop thru the array
  for(int z = 0; z < SIZE_Z; ++z) {  // loop thru z

    for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y

      for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

        for(int c = 0; c < SIZE_C; ++c) {  // loop thru c

          // define the thresholds
          ca->shapes[x][y][z].cells[c].threshold = 0.5;

          // make sure all of they are initialized with zeroes to prevent nan's
          ca->shapes[x][y][z].cells[c].sum = ((float)rand()) / ((float)RAND_MAX);
          ca->shapes[x][y][z].cells[c].storage = ((float)rand()) / ((float)RAND_MAX);
          ca->shapes[x][y][z].cells[c].output = ((float)rand()) / ((float)RAND_MAX);

          for(int w = 0; w < SIZE_W; ++w) {  // loop thru w

            // define the weights
           ca->shapes[x][y][z].cells[c].weights[w] = 
              ((float)rand()) / ((float)RAND_MAX) - 0.5;

           // define the deltas
           ca->shapes[x][y][z].cells[c].deltas[w] = 
              ca->temperature * ((float)rand()) / ((float)RAND_MAX) - 0.5;

          }

        }

      }

    }

  }

}



// define the min function
int min(int a, int b) {

  if (a >= b) {

    return b;

  } else {

    return a;

  }

}



// define the max function
int max(int a, int b) {

  if (a <= b) {

    return b;

  } else {

    return a;

  }

}



// generate deltas and add them to the cognitive array's weights
void deltas(struct cognitiveArray *ca, int mode) {

  // randomly generate deltas
  // loop thru the array
  for(int z = 0; z < SIZE_Z; ++z) {  // loop thru z

    for(int y = 0; y < SIZE_Y; ++y) {  // loop thru y

      for(int x = 0; x < SIZE_X; ++x) {  // loop thru x

        for(int c = 0; c < SIZE_C; ++c) {  // loop thru c

          for(int w = 0; w < SIZE_W; ++w) {  // loop thru w

            // add the deltas to the weights accoding to command from fon() evaluation done last iteration
            ca->shapes[x][y][z].cells[c].weights[w] += 
              ca->command * ca->shapes[x][y][z].cells[c].deltas[w]; 
            
            // generate the deltas
            ca->shapes[x][y][z].cells[c].deltas[w] = 
              ca->temperature * ((float)rand()) / ((float)RAND_MAX);

            // add the new deltas to the weights
            ca->shapes[x][y][z].cells[c].weights[w] += 
              ca->command * ca->shapes[x][y][z].cells[c].deltas[w]; 

          }

        }

      }

    }

  }

}



// evaluate figure of merit
void fom(struct cognitiveArray *ca, int mode) {

  // calculate a desired activity level
  float level = SIZE_X * SIZE_Y * SIZE_Z * SIZE_C;

  // this will eventually be more meaningful such as human interaction or prediction, etc. 
  // for now, however, we will use the activity level as fom.  
  ca->command = 1.0;

  if (fabs(ca->activity) > level) {

    ca->command = -1.0;

  }

}



// use a sigmoid function to keep things from exploding (hopefully)
void sigmoid(struct sigmoid *s) {

  s->y = 1.0 / (s->shape + exp(-s->x / s->tau) ); 

}

















